"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import ast
import collections
import datetime
import os
import random
import re
import site
import string
import subprocess

from CP3SlurmUtils import SpellChecker
from CP3SlurmUtils.Exceptions import ConfigurationException
from CP3SlurmUtils.Exceptions import CP3SlurmUtilsException

try:
    import configparser
except ImportError:
    import ConfigParser as configparser
defaultsCfg = configparser.ConfigParser()
defaultsCfg.add_section('USER_CONFIG_DEFAULTS')
defaultsCfg.set('USER_CONFIG_DEFAULTS', 'sbatch_partition', 'cp3')
defaultsCfg.set('USER_CONFIG_DEFAULTS', 'sbatch_qos', 'cp3')
defaultsCfg.set('USER_CONFIG_DEFAULTS', 'sbatch_time', '0-04:00')
defaultsCfg.set('USER_CONFIG_DEFAULTS', 'sbatch_memPerCPU', '2048')
defaultsCfg.set('USER_CONFIG_DEFAULTS', 'scratchDir', '${LOCALSCRATCH}')
defaultsCfg.add_section('USER_CONFIG_VALID_VALUES')
defaultsCfg.set('USER_CONFIG_VALID_VALUES', 'sbatch_partition', 'Def,cp3,cp3-fast,cp3-gpu')
defaultsCfg.set('USER_CONFIG_VALID_VALUES', 'sbatch_qos', 'normal,cp3,cp3-gpu')
defaultsCfg.set('USER_CONFIG_VALID_VALUES', 'environmentType', 'cms')
defaultsCfg.add_section('USER_CONFIG_PARTITIONS_FOR_QOS')
defaultsCfg.set('USER_CONFIG_PARTITIONS_FOR_QOS', 'qos_cp3', 'cp3,cp3-fast')
defaultsCfg.set('USER_CONFIG_PARTITIONS_FOR_QOS', 'qos_cp3-gpu', 'cp3-gpu')
defaultsCfg.set('USER_CONFIG_PARTITIONS_FOR_QOS', 'alternative_qos', 'normal')
xdgConfigDirsDef = os.path.join(os.path.sep, 'etc', 'xdg')
xdgConfigDirs = os.getenv('XDG_CONFIG_DIRS', xdgConfigDirsDef)
if xdgConfigDirs.strip() == '':
    xdgConfigDirs = xdgConfigDirsDef
xdgConfigHomeDef = os.path.join(os.path.expanduser('~'), '.config')
xdgConfigHome = os.getenv('XDG_CONFIG_HOME', xdgConfigHomeDef)
if xdgConfigHome.strip() == '':
    xdgConfigHome = xdgConfigHomeDef
siteUserBase = site.USER_BASE
virtualenv = os.getenv('VIRTUAL_ENV')
defaultsCfgLocations = {
    'system'     : [os.path.join(os.path.sep, 'etc', 'CP3SlurmUtils', 'defaults.cfg')] + \
                   [os.path.join(xdgConfigDir.strip(), 'CP3SlurmUtils', 'defaults.cfg') for xdgConfigDir in xdgConfigDirs.split(':') if xdgConfigDir.strip()],
    'local'      : [os.path.join(siteUserBase, 'etc', 'CP3SlurmUtils', 'defaults.cfg')],
    'user'       : [os.path.join(xdgConfigHome, 'CP3SlurmUtils', 'defaults.cfg')],
    'virtualenv' : [os.path.join(virtualenv, 'etc', 'CP3SlurmUtils', 'defaults.cfg')] if virtualenv else [],
}
defaultsCfg.read(defaultsCfgLocations['system'] + defaultsCfgLocations['local'] + defaultsCfgLocations['user'] + defaultsCfgLocations['virtualenv'])

timestamp = str(datetime.datetime.now().strftime("%Y%m%d-%H%M%S%f"))
randomstr = ''.join(random.choice(string.ascii_lowercase) for i in range(4))


validConfigParams = collections.OrderedDict({
    'sbatch_partition'            : {'mandatory': False, 'type': str,                        'default': defaultsCfg.get('USER_CONFIG_DEFAULTS', 'sbatch_partition'), 'sbatch_option_long': 'partition',   'sbatch_option_short': 'p'},
    'sbatch_qos'                  : {'mandatory': False, 'type': str,                        'default': defaultsCfg.get('USER_CONFIG_DEFAULTS', 'sbatch_qos'),       'sbatch_option_long': 'qos',         'sbatch_option_short': None},
    'sbatch_time'                 : {'mandatory': False, 'type': str,                        'default': defaultsCfg.get('USER_CONFIG_DEFAULTS', 'sbatch_time'),      'sbatch_option_long': 'time',        'sbatch_option_short': 't'},
    'sbatch_memPerCPU'            : {'mandatory': False, 'type': str,                        'default': defaultsCfg.get('USER_CONFIG_DEFAULTS', 'sbatch_memPerCPU'), 'sbatch_option_long': 'mem-per-cpu', 'sbatch_option_short': None},
    'sbatch_chdir'                : {'mandatory': False, 'type': str,                        'default': '.',                                                         'sbatch_option_long': 'chdir',       'sbatch_option_short': 'D'},
    'sbatch_output'               : {'mandatory': False, 'type': str,                        'default': '',                                                          'sbatch_option_long': 'output',      'sbatch_option_short': 'o'},
    'sbatch_error'                : {'mandatory': False, 'type': str,                        'default': '',                                                          'sbatch_option_long': 'error',       'sbatch_option_short': 'e'},
    'sbatch_additionalOptions'    : {'mandatory': False, 'type': list, 'element_type': str,  'default': []},
    'scratchDir'                  : {'mandatory': False, 'type': str,                        'default': defaultsCfg.get('USER_CONFIG_DEFAULTS', 'scratchDir')},
    'handleScratch'               : {'mandatory': False, 'type': bool,                       'default': False},
    'environmentType'             : {'mandatory': False, 'type': str,                        'default': ''},
    'cmsswDir'                    : {'mandatory': False, 'type': str,                        'default': ''},
    'inputSandboxContent'         : {'mandatory': False, 'type': list, 'element_type': str,  'default': []},
    'inputSandboxDir'             : {'mandatory': False, 'type': str,                        'default': '<config.sbatch_chdir>'},
    'inputSandboxFilename'        : {'mandatory': False, 'type': str,                        'default': 'input_sandbox_{}_{}.tar.gz'.format(timestamp, randomstr)},
    'batchScriptsDir'             : {'mandatory': False, 'type': str,                        'default': '<config.sbatch_chdir>'},
    'batchScriptsFilename'        : {'mandatory': False, 'type': str,                        'default': 'slurm_batch_script_{}_{}.sh'.format(timestamp, randomstr)},
    'stageout'                    : {'mandatory': False, 'type': bool,                       'default': True},
    'stageoutFiles'               : {'mandatory': False, 'type': list, 'element_type': str,  'default': []},
    'stageoutDir'                 : {'mandatory': False, 'type': str,                        'default': '<config.sbatch_chdir>'},
    'writeLogsOnWN'               : {'mandatory': False, 'type': bool,                       'default': True},
    'separateStdoutStderrLogs'    : {'mandatory': False, 'type': bool,                       'default': False},
    'stdoutFilename'              : {'mandatory': False, 'type': str,                        'default': ''},
    'stderrFilename'              : {'mandatory': False, 'type': str,                        'default': ''},
    'stageoutLogs'                : {'mandatory': False, 'type': bool,                       'default': True},
    'stageoutLogsDir'             : {'mandatory': False, 'type': str,                        'default': '<config.stageoutDir>/logs'},
    'useJobArray'                 : {'mandatory': False, 'type': bool,                       'default': True},
    'maxRunningJobs'              : {'mandatory': False, 'type': int,                        'default': None},
    'numJobs'                     : {'mandatory': False, 'type': int,                        'default': None},
    'inputParamsNames'            : {'mandatory': False, 'type': list, 'element_type': str,  'default': []},
    'inputParams'                 : {'mandatory': False, 'type': list, 'element_type': list, 'default': []},
    'doCheckpointing'             : {'mandatory': False, 'type': bool,                       'default': False},
    'checkpointSignal'            : {'mandatory': False, 'type': str,                        'default': ""},
    'checkpointSecsBeforeTimeout' : {'mandatory': False, 'type': int,                        'default': 0},
    'payload'                     : {'mandatory': True,  'type': str,                        'default': ""},
})

renamedParams = [
]

deprecatedParams = [
]

invalidConfigMsg = "ERROR: Invalid Configuration:"


def strip(string, separator=''):
    if separator:
        return separator.join([s.strip() for s in string.split(separator) if s.strip()])
    else:
        return string.strip()


def fixConfigDefaults(config):
    """_fixConfigDefaults_
    Change the default values of some configuration parameters
    depending on the actual values of other configuration parameters.
    """
    if getattr(config, 'writeLogsOnWN', validConfigParams['writeLogsOnWN']['default']):
        validConfigParams['sbatch_output']['default'] = "/dev/null"
        validConfigParams['sbatch_error']['default'] = "/dev/null"
    if getattr(config, 'useJobArray', validConfigParams['useJobArray']['default']): 
        validConfigParams['stdoutFilename']['default'] = "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out"
        validConfigParams['stderrFilename']['default'] = "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err"
    else:
        validConfigParams['stdoutFilename']['default'] = "slurm-${SLURM_JOB_ID}.out"
        validConfigParams['stderrFilename']['default'] = "slurm-${SLURM_JOB_ID}.err"


def setConfigDefaults(config):
    """_setConfigDefaults_
    Apply the default values to the optional attributes that are not
    already set.
    """
    for paramName, paramSpecs in list(validConfigParams.items()):
        if paramSpecs['mandatory']:
            continue
        if getattr(config, paramName, None) is None or (paramSpecs['type'] == str and getattr(config, paramName) == ''):
            refParamName = paramName
            refParamSpecs = paramSpecs
            n = 0
            while (getattr(config, refParamName, None) is None or (refParamSpecs['type'] == str and getattr(config, refParamName) == '')) and \
                  (isinstance(refParamSpecs['default'], str) and '<' in refParamSpecs['default'] and '>' in refParamSpecs['default']):
                refParamName, postStr = refParamSpecs['default'].split('<config.')[1].split('>')
                preStr = refParamSpecs['default'].split('<config.')[0]
                if refParamName not in validConfigParams:
                    msg = "ERROR: Default value for parameter '{param}' is referencing an invalid parameter '{reference_param}'."
                    msg = msg.format(param=paramName, reference_param=refParamName)
                    raise CP3SlurmUtilsException(msg)
                refParamSpecs = validConfigParams[refParamName]
                n += 1
                if n > len(validConfigParams):
                    msg = "ERROR: Seems parameter '{param}' will not be able to get a default value."
                    msg += "\nLooped already {0} times, which is more than the number of available configuration parameters ({1})."
                    msg = msg.format(n, len(validConfigParams), param=paramName)
                    raise CP3SlurmUtilsException(msg)
            if refParamName != paramName:
                if isinstance(getattr(config, refParamName, refParamSpecs['default']), str):
                    setattr(config, paramName, preStr + getattr(config, refParamName, refParamSpecs['default']) + postStr)
                else:
                    setattr(config, paramName, getattr(config, refParamName, refParamSpecs['default']))
            else:
                setattr(config, paramName, paramSpecs['default'])


def validateConfigBeforeDefaults(config):
    """_validateConfigBeforeDefaults_
    Do a validation of a Configuration before any defauls are appield.
    """
    validateConfigBasic(config)
    validateConfigDirs(config)


def checkType(attr, requiredType):
    """_checkType_
    Check the type of an attribute.
    """
    if isinstance(attr, bool) and requiredType == int:
        return False
    return isinstance(attr, requiredType)


def validateConfigBasic(config):
    """_validateConfigBasic_
    Do a basic validation of a Configuration.
    """
    # Some parameters may have been renamed. Check if there is an old
    # parameter and tell the user what is the new parameter name.
    for renamedParam in renamedParams:
        if hasattr(config, renamedParam['oldName']):
            msg = invalidConfigMsg
            msg += "\nParameter '{param}' has been renamed to '{new_param}'. Please change your configuration file accordingly."
            msg = msg.format(param=renamedParam['oldName'], new_param=renamedParam['newName'])
            raise ConfigurationException(msg)
    # Check if there is a deprecated parameter.
    for deprecatedParam in deprecatedParams:
        if hasattr(config, deprecatedParam['name']):
            msg = invalidConfigMsg
            msg += "\nParameter '{param}' has been deprecated in version {version}."
            if deprecatedParam['msg']:
                msg += "\n{message}"
            msg = msg.format(param=deprecatedParam['name'], version=deprecatedParam['version'], message=deprecatedParam['msg'])
            raise ConfigurationException(msg)
    # Check the type of the existing parameters.
    for paramName, paramSpecs in list(validConfigParams.items()):
        if getattr(config, paramName, None) is None:
            continue
        attr = getattr(config, paramName)
        if not checkType(attr, paramSpecs['type']):
            msg = invalidConfigMsg
            msg += "\nParameter '{param}' is not of type '{type}'."
            msg = msg.format(param=paramName, type=paramSpecs['type'])
            raise ConfigurationException(msg)
        # Remove leading and trailing spaces in configuration parameters that are strings.
        if paramSpecs['type'] == str:
            if paramName == 'sbatch_partition':
                setattr(config, paramName, strip(attr, ','))
            else:
                setattr(config, paramName, strip(attr))
        if paramSpecs['type'] == list:
            for i, el in enumerate(attr):
                if not checkType(el, paramSpecs['element_type']):
                    msg = invalidConfigMsg
                    msg += "\nList element {0} in parameter '{param}' is not of type '{type}'."
                    msg = msg.format(i, param=paramName, type=paramSpecs['element_type'])
                    raise ConfigurationException(msg)
            # Remove leading and trailing spaces in configuration parameters that are lists of strings.
            # Remove also empty strings.
            if paramSpecs['element_type'] == str:
                setattr(config, paramName, [s for s in list(map(str.strip, attr)) if s])
    # Check if there is an unknown parameter and try to suggest the correct parameter name.
    SpellChecker.DICTIONARY = SpellChecker.train(list(validConfigParams.keys()))
    for param in config.getConfigAttrs_():
        if not SpellChecker.is_correct(param):
            msg = invalidConfigMsg
            msg += "\nParameter '{param}' is not known."
            if SpellChecker.correct(param) != param:
                msg += " Maybe you mean '{suggested_param}'?"
            msg = msg.format(param=param, suggested_param=SpellChecker.correct(param))
            raise ConfigurationException(msg)
    checkMandatoryConfigParams(config)


def checkMandatoryConfigParams(config, params=None):
    """_checkMandatoryConfigParams_
    Check the mandatory parameters of a Configuration.
    params: a sublist of parameters to check.
    """
    # Mandatory parameters should be there and can not have an empty value.
    for paramName, paramSpecs in list(validConfigParams.items()):
        if params and paramName not in params:
            continue
        if paramSpecs['mandatory']:
            if not hasattr(config, paramName):
                msg = invalidConfigMsg
                msg += "\nParameter '{param}' is mandatory."
                msg = msg.format(param=paramName)
                raise ConfigurationException(msg)
            elif getattr(config, paramName) in [None, [], {}] or (paramSpecs['type'] == str and getattr(config, paramName) == ''):
                msg = invalidConfigMsg
                msg += "\nParameter '{param}' is mandatory (and can not be defined as None or an empty string/list/dict)."
                msg = msg.format(param=paramName)
                raise ConfigurationException(msg)


def validateConfigDirs(config):
    """_validateConfigDirs_
    Validate the directories of a Configuration checking that they
    do not point to an existing file.
    """
    for paramName, paramSpecs in list(validConfigParams.items()):
        if not paramName.endswith('Dir'):
            continue
        if getattr(config, paramName, None) is None:
            continue
        dirName = getattr(config, paramName)
        if os.path.isfile(dirName):
            msg = invalidConfigMsg
            msg += "\nThe directory provided in parameter '{param}' is already an existing file: {value}"
            msg = msg.format(param=paramName, value=dirName)
            raise ConfigurationException(msg)


def validateConfigAfterDefaults(config):
    """_validateConfigAfterDefaults_
    Do a basic validation of a Configuration assuming defaults were applied.
    """
    paramNotDefinedMsg = invalidConfigMsg+"\nParameter '{param}' is not defined."
    paramInvalidValueMsg = invalidConfigMsg+"\nParameter '{param}' has an invalid value: '{value}'."
    paramInvalidValueMsgSuggest = paramInvalidValueMsg+"\nValid values are: '{valid}'."
    # Check that there are no sbatch options specified in 'sbatch_additionalOptions'
    # for which we already have a configuration parameter available.
    for paramName, paramSpecs in list(validConfigParams.items()):
        if paramName.startswith('sbatch_') and paramName != 'sbatch_additionalOptions':
            sbatchOptLong = paramSpecs['sbatch_option_long']
            sbatchOptShort = paramSpecs['sbatch_option_short']
            sbatchAddOpts = [opt.split('=')[0].strip() for opt in config.sbatch_additionalOptions]
            if ('--{}'.format(sbatchOptLong) in sbatchAddOpts):
                msg = invalidConfigMsg
                msg += "\nPlease use the configuration parameter '{param}'"
                msg += " instead of setting the --{long_option} sbatch option in 'sbatch_additionalOptions'."
                msg = msg.format(param=paramName, long_option=sbatchOptLong)
                raise ConfigurationException(msg)
            if sbatchOptShort and ('-{}'.format(sbatchOptShort) in sbatchAddOpts):
                msg = invalidConfigMsg
                msg += "\nPlease use the configuration parameter '{param}'"
                msg += " instead of setting the -{short_option} sbatch option in 'sbatch_additionalOptions'."
                msg = msg.format(param=paramName, short_option=sbatchOptShort)
                raise ConfigurationException(msg)
    # Check that time, memory and partition are defined.
    # We do not make these parameters mandatory in validConfigParams,
    # because in principle there are default values for them.
    # But since the default values can be omitted in defaults.cfg,
    # we have to make sure that these parameters are defined.
    if not config.sbatch_time:
        msg = paramNotDefinedMsg.format(param='sbatch_time')
        raise ConfigurationException(msg)
    if not config.sbatch_memPerCPU:
        msg = paramNotDefinedMsg.format(param='sbatch_memPerCPU')
        raise ConfigurationException(msg)
    if not config.sbatch_partition:
        msg = paramNotDefinedMsg.format(param='sbatch_partition')
        raise ConfigurationException(msg)
    # Check the partition.
    validPartitions = [ x for x in strip(defaultsCfg.get('USER_CONFIG_VALID_VALUES', 'sbatch_partition'), ',').split(',') if x ]
    if validPartitions:
        for partition in config.sbatch_partition.split(','):
            if partition not in validPartitions:
                msg = paramInvalidValueMsg
                msg += "\nIt should be a string of comma separated cluster partitions. Acceptable partitions are: '{valid}'."
                msg = msg.format(param='sbatch_partition', value=config.sbatch_partition, valid="', '".join(validPartitions))
                raise ConfigurationException(msg)
    # Check the qos.
    if config.sbatch_qos:
        validQOSs = [ x for x in strip(defaultsCfg.get('USER_CONFIG_VALID_VALUES', 'sbatch_qos'), ',').split(',') if x ]
        if validQOSs:
            if config.sbatch_qos not in validQOSs:
                msg = paramInvalidValueMsgSuggest.format(param='sbatch_qos', value=config.sbatch_qos, valid="', '".join(validQOSs))
                raise ConfigurationException(msg)
        # Check the compatibility between qos and partition.
        if defaultsCfg.has_option('USER_CONFIG_PARTITIONS_FOR_QOS', 'qos_'+config.sbatch_qos):
            partitionsForQOS = [ x for x in strip(defaultsCfg.get('USER_CONFIG_PARTITIONS_FOR_QOS', 'qos_'+config.sbatch_qos), ',').split(',') if x ]
            if partitionsForQOS:
                for partition in config.sbatch_partition.split(','):
                    if partition not in partitionsForQOS:
                        msg = invalidConfigMsg
                        msg += "\nParameters 'sbatch_partition' and 'sbatch_qos' have incompatible values:"
                        msg += "\nsbatch_partition: '{partition}'"
                        msg += "\nsbatch_qos: '{qos}'"
                        msg += "\nQuality of service '{qos}' is forbidden on cluster partitions other than '{partitions_for_qos}' (jobs would be accepted anyway, but would never start running)."
                        alternativeQOS = defaultsCfg.get('USER_CONFIG_PARTITIONS_FOR_QOS', 'alternative_qos')
                        if alternativeQOS:
                            msg += "\nThus, when submitting jobs to a cluster partition that is not in the above list, the quality of service should be set to '{alternative_qos}'."
                        msg = msg.format(partition=config.sbatch_partition, qos=config.sbatch_qos, partitions_for_qos="', '".join(partitionsForQOS), alternative_qos=alternativeQOS)
                        raise ConfigurationException(msg)
    # Check the environmentType.
    if config.environmentType:
        validEnvTypes = [ x for x in strip(defaultsCfg.get('USER_CONFIG_VALID_VALUES', 'environmentType'), ',').split(',') if x ]
        if validEnvTypes:
            if config.environmentType.lower() not in validEnvTypes:
                msg = paramInvalidValueMsg
                msg += "\nValid values are (apart from the empty string or None): '{valid}'."
                msg = msg.format(param='environmentType', value=config.environmentType, valid="', '".join(validEnvTypes))
                raise ConfigurationException(msg)
        # For CMS jobs, check that a CMSSW work area was given, that it exists and is a directory.
        if config.environmentType.lower() == 'cms':
            if not config.cmsswDir:
                msg = invalidConfigMsg
                msg += "\nParameter 'environmentType' is set to 'cms', but parameter 'cmsswDir' was not specified."
                msg += "\nTo setup the CMS environment, a CMSSW work area has to be specified where to run the cmsenv command."
                msg += "\nIf you don't need to setup the CMS environment, set 'environmentType' to an empty string or None."
                raise ConfigurationException(msg)
            if not os.path.isdir(config.cmsswDir):
                msg = invalidConfigMsg
                msg += "\nParameter '{param}' does not correspond to an existing directory: '{value}'."
                msg = msg.format(param='cmsswDir', value=config.cmsswDir)
                raise ConfigurationException(msg)
    # Check that at least one of 'numJobs' or 'inputParams' was specified.
    if config.numJobs is None and len(config.inputParams) == 0:
        msg = invalidConfigMsg
        msg += "\nAt least one of 'numJobs' or 'inputParams' must be specified (with a sensitive value)."
        raise ConfigurationException(msg)
    # Check that 'numJobs' is either None or a positive integer.
    # If positive, and if there are input parameters for the jobs,
    # check that there are enough input parameters for all jobs.
    if config.numJobs is not None:
        if config.numJobs <= 0:
            msg = invalidConfigMsg
            msg += "\nParameter 'numJobs' must be a positive integer (or None)."
            raise ConfigurationException(msg)
        if 0 < len(config.inputParams) and len(config.inputParams) < config.numJobs:
            msg = invalidConfigMsg
            msg += "\nThe length of the 'inputParams' list ({0}) is smaller than 'numJobs' ({1})."
            msg = msg.format(len(config.inputParams), config.numJobs)
            raise ConfigurationException(msg)
    # Check that 'maxRunningJobs' is either None or a non-negative integer.
    if config.maxRunningJobs is not None:
        if config.maxRunningJobs < 0:
            msg = invalidConfigMsg
            msg += "\nParameter 'maxRunningJobs' must be a non-negative integer (or None)."
            raise ConfigurationException(msg)
    # Check the consistency in length between the lists 'inputParamsNames' and 'inputParams'.
    if config.inputParamsNames and not config.inputParams:
        msg = invalidConfigMsg
        msg += "\nInconsistency between 'inputParamsNames' and 'inputParams'."
        msg += "\nYou provided 'inputParamsNames', but not 'inputParams'."
        msg += "\nEither both of them have to be provided or none of them."
        raise ConfigurationException(msg)
    if not config.inputParamsNames and config.inputParams:
        msg = invalidConfigMsg
        msg += "\nInconsistency between 'inputParamsNames' and 'inputParams'."
        msg += "\nYou provided 'inputParams', but not 'inputParamsNames'."
        msg += "\nEither both of them have to be provided or none of them."
        raise ConfigurationException(msg)
    numJobInputParams = len(config.inputParamsNames)
    if numJobInputParams > 0:
        for i, jobInputParams in enumerate(config.inputParams):
            if len(jobInputParams) != numJobInputParams:
                msg = invalidConfigMsg
                msg += "\nInconsistency between 'inputParamsNames' and 'inputParams'."
                msg += "\nThe list element {0} of the 'inputParams' list contains {1} parameters"
                msg += " while according to the list 'inputParamsNames' there should be {2} parameters per job."
                msg = msg.format(i, len(jobInputParams), numJobInputParams)
                raise ConfigurationException(msg)
    # If 'stageout' is True, and if there are user files to stageout,
    # then 'stageoutDir' must be specified.
    if config.stageout and config.stageoutFiles and not config.stageoutDir:
        msg = invalidConfigMsg
        msg += "\nParameter 'stageout' is True, but parameter 'stageoutDir' was not specified (while there are -a priori- files to stageout)."
        raise ConfigurationException(msg)
    # If 'writeLogsOnWN' is True, there should be no sbatch --output or --error option specified.
    if config.writeLogsOnWN:
        if config.sbatch_output != '/dev/null' or config.sbatch_error != '/dev/null': 
            msg = invalidConfigMsg
            msg += "\nParameters 'writeLogsOnWN' and 'sbatch_output'/'sbatch_error' are mutually exclusive."
            msg += "\nNote: You either write the logs on the WN (and eventually stage them out)"
            msg += " or you use the 'sbatch_output'/'sbatch_error' options (eventually leaving them empty so to use slurm defaults)"
            msg += " to let the logs be created outside the WN."
            raise ConfigurationException(msg)
    # Make a sanity check of the input sandbox, batch script and logs filenames.
    paramInvalidRegexMsg = paramInvalidValueMsg + "\nIt should contain only the following characters: a-zA-Z0-9${{}}._-"
    if config.inputSandboxFilename:
        if not re.match("^[a-zA-Z0-9${}._-]+$", config.inputSandboxFilename):
            msg = paramInvalidRegexMsg.format(param='inputSandboxFilename', value=config.inputSandboxFilename)
            raise ConfigurationException(msg)
    if config.batchScriptsFilename:
        if not re.match("^[a-zA-Z0-9${}._-]+$", config.batchScriptsFilename):
            msg = paramInvalidRegexMsg.format(param='batchScriptsFilename', value=config.batchScriptsFilename)
            raise ConfigurationException(msg)
    if config.stdoutFilename:
        if not re.match("^[a-zA-Z0-9${}._-]+$", config.stdoutFilename):
            msg = paramInvalidRegexMsg.format(param='stdoutFilename', value=config.stdoutFilename)
            raise ConfigurationException(msg)
    if config.stderrFilename:
        if not re.match("^[a-zA-Z0-9${}._-]+$", config.stderrFilename):
            msg = paramInvalidRegexMsg.format(param='stderrFilename', value=config.stderrFilename)
            raise ConfigurationException(msg)
    # If 'stageoutLogs' is True, then 'writeLogsOnWN' must also be True.
    if config.stageoutLogs and not config.writeLogsOnWN:
        msg = invalidConfigMsg
        msg += "\nParameter 'stageoutLogs' is True, but parameter 'writeLogsOnWN' is False."
        raise ConfigurationException(msg)
    # If 'stageoutLogs' is True, then 'stageoutLogsDir' must be specified.
    if config.stageoutLogs and not config.stageoutLogsDir:
        msg = invalidConfigMsg
        msg += "\nParameter 'stageoutLogs' is True, but parameter 'stageoutLogsDir' was not specified."
        raise ConfigurationException(msg)
    # For checkpointing, accept only SIGUSR1/2 signals.
    if config.doCheckpointing:
        if config.checkpointSignal not in ['SIGUSR1', 'SIGUSR2', '']:
            msg = "ERROR: Invalid Configuration:"
            msg += "\nParameter 'checkpointSignal' has an invalid value: %s" % (config.checkpointSignal)
            msg += "\nAllowed values are: SIGUSR1, SIGUSR2."
            msg += "\nIt can also be set to an empty string to disable checkpoint-and-exit under a system signal."
            raise ConfigurationException(msg)
    # For checkpointing, set negative intervals in seconds to 0.
    if config.doCheckpointing:
        setattr(config, 'checkpointSecsBeforeTimeout', max(getattr(config, 'checkpointSecsBeforeTimeout'), 0))
    # If 'checkpointSecsBeforeTimeout' is given, then 'checkpointSignal' should be specified.
    if config.doCheckpointing:
        if config.checkpointSecsBeforeTimeout > 0:
            if not config.checkpointSignal:
                msg = "ERROR: Invalid Configuration:"
                msg += "\nParameter 'checkpointSecsBeforeTimeout' is set (greater than 0), but parameter 'checkpointSignal' was not specified."
                raise ConfigurationException(msg)


def fixConfigDirs(config):
    """_fixConfigDirs_
    Replace relative paths by absolute paths in a Configuration.
    Expand ~.
    """
    for paramName, paramSpecs in list(validConfigParams.items()):
        if not paramName.endswith('Dir'):
            continue
        if getattr(config, paramName, None) is None:
            continue
        dirName = getattr(config, paramName)
        if dirName and not dirName.startswith('$'):
            dirName = os.path.expanduser(dirName)
            dirName = os.path.abspath(dirName)
        setattr(config, paramName, dirName)


def createConfigDirs(config):
    """_createConfigDirs_
    Create the directories of a Configuration.
    """
    for paramName, paramSpecs in list(validConfigParams.items()):
        if paramName not in ['inputSandboxDir', 'batchScriptsDir']:
            continue
        if not getattr(config, paramName, None):
            continue
        returncode = subprocess.call(['mkdir', '-p', getattr(config, paramName)])
        if returncode != 0:
            msg = "ERROR: Failed to create directory {}".format(getattr(config, paramName))
            raise CP3SlurmUtilsException(msg)



from CP3SlurmUtils.Configuration import Configuration

config = Configuration()

#--------------------------------------------------------------------------------
# 1. SLURM sbatch command options
#--------------------------------------------------------------------------------

# The cluster partition(s) where jobs can run.
# Comma-separated list (a comma means a union) of the following allowed values:
# "Def" - refers to the CISM portion of the cluster;
# "cp3" - refers to the CP3 portion of the cluster.
# CP3 users are allowed to run their jobs on the whole cluster. The same is true for non-CP3 users.
# The difference is in the priorities: CP3 users can request a special quality of service (see 'sbatch_qos' below)
# on the CP3 partition, while non-CP3 users can not. That's why CP3 users are recommended to run
# only on the CP3 partition. But if you want to submit your jobs to the CISM partition
# (e.g. because the CP3 partition is full while the CISM partition is not), you can do so by setting
# this parameter to "Def" (or "Def,cp3" to submit to the whole cluster).
# The default value of this parameter can be configured in the defaults.cfg file.
# At CP3 the default is "cp3".
# This parameter has precedence over the SLURM environment variable SBATCH_PARTITION.
#config.sbatch_partition = "cp3"

# This parameter is used to request a given quality of service for the jobs.
# Valid values are "normal" and "cp3". The quality of service "cp3" is only allowed in the CP3 partition
# and only for CP3 users, and what it does is to boost the priority by 10000 (compare this to the maximum
# priority of 2500 in the "normal" quality of service). Thus, the recommendation for CP3 users is to run
# on the CP3 partition with quality of service "cp3".
# If 'sbatch_partition' is not set to "cp3", then 'sbatch_qos' should be set to "normal".
# The default value of this parameter can be configured in the defaults.cfg file.
# At CP3 the default is "cp3" if 'sbatch_partition' is set to "cp3"; otherwise the default is "normal".
# This parameter has precedence over the SLURM environment variable SBATCH_QOS.
#config.sbatch_qos = "cp3"

# The working directory where the SLURM batch script will be executed.
# Can be specified with a full path or with a relative path.
# In the last case, SLURM will consider it relative to the directory where the sbatch command is executed.
# If no other location is specified for the SLURM stdout and stderr files,
# this working directory will be the directory where these files will be written when a job starts running.
# This directory must exist already.
# Defaults to the current working directory (symlinks are resolved to physical paths).
#config.sbatch_chdir = "."

# Maximum cpu time needed by each job (days-hours:minutes).
# CMS jobs should tipically not run for more than 4 hours. But if you need more, you can (and should) request it.
# SLURM will kill the jobs that use more cpu time than requested.
# Having a good estimate of the maximum cpu time your jobs need may help your jobs to start running earlier.
# In the CP3 cluster, SLURM has a backfill scheduling mechanism that will start lower priority jobs if doing so does not delay the expected 
# start time of any higher priority jobs. Also, since the expected start time of pending jobs depends upon the expected
# completion time of running jobs, reasonably accurate time limits are important for backfill scheduling to work well.
# The default value of this parameter can be configured in the defaults.cfg file.
# At CP3 the default is "0-04:00" (4 hours).
#config.sbatch_time = "0-04:00"

# Maximum memory per cpu needed by each job (in MB).
# CMS jobs should not use more than 2 GB. But if you imperatively need more, you can (and should) request it.
# In the CP3 cluster, SLURM will kill the jobs that use more memory (RSS) than requested.
# Obviously, the more you request the longer your jobs will wait in the queue for the resources to become available.
# The default value of this parameter can be configured in the defaults.cfg file.
# At CP3 the default is "2048" (2 GB).
#config.sbatch_memPerCPU = "2048"

# SLURM will connect the batch script's standard output directly to the file specified in this parameter.
# By default both standard output and standard error are directed to the same file.
# For job arrays, the default filename is "slurm-%A_%a.out", where "%A" is replaced by the job ID and "%a" with the array index.
# For other jobs, the default filename is "slurm-%j.out", where "%j" is replaced by the job ID.
# If a relative path is given, it will be relative to the work directory 'sbatch_chdir'.
# If 'writeLogsOnWN' is True (the default), this parameter will default to "/dev/null"
# so that SLURM does not create log files by itself outside the worker nodes. Moreover, assigning
# to it other value than "/dev/null" will produce an error.
# Setting this parameter to None or to an empty string will not turn off the SLURM output logging;
# instead it will simply let SLURM use the default file for the standard output.
# Defaults to an empty string.
#config.sbatch_output = ""

# SLURM will connect the batch script's standard error directly to the file specified in this parameter. 
# By default both standard output and standard error are directed to the same file.
# For job arrays, the default filename is "slurm-%A_%a.out", where "%A" is replaced by the job ID and "%a" with the array index.
# For other jobs, the default filename is "slurm-%j.out", where "%j" is replaced by the job ID.
# If a relative path is given, it will be relative to the work directory 'sbatch_chdir'.
# If 'writeLogsOnWN' is True (the default), this parameter will default to "/dev/null"
# so that SLURM does not create log files by itself outside the worker nodes. Moreover, assigning
# to it other value than "/dev/null" will produce an error.
# Setting this parameter to None or to an empty string will not turn off the SLURM error logging;
# instead it will simply let SLURM use the default file for the standard error.
# Defaults to an empty string.
#config.sbatch_error = ""

# Use this parameter to set any other SLURM sbatch options you may need.
# Example: config.sbatch_additionalOptions = ["--mail-type=END", "--mail-user=<your-email-address>"]
# Defaults to an empty list (no additional options).
#config.sbatch_additionalOptions = []

#--------------------------------------------------------------------------------
# 2. User batch script parameters that are same for all jobs
#--------------------------------------------------------------------------------

# This parameter specifies a scratch directory where the job should run.
# It could well be defined via an environment variable: config.scratchDir = "${scratchDir}".
# In that case, the environment variable must exist during job execution.
# At CP3, each job has a scratch directory that is defined, created and cleaned-up by SLURM;
# the scratch directory name is defined after job submission and is exported to an environment
# variable named LOCALSCRATCH.
# If you don't want to use a scratch directory, set this parameter to an empty string or None.
# The default value of this parameter can be configured in the defaults.cfg file.
# At CP3 the default is "${LOCALSCRATCH}".
#config.scratchDir = "${LOCALSCRATCH}"

# Set this parameter to True to create/remove the scratch directory before/after
# the user payload starts/ends.
# Defaults to False.
#config.handleScratch = False

# This parameter specifies the environment that has to be setup before the payload.
# For the moment the only possible environment setup is the CMS one, in which case
# a CMSSW work area should also be specified in the 'cmsswDir' parameter below.
# Valid values are: "cms" (case insensitive).
# To not setup any environment, set this parameter to an empty string or None.
# Defaults to an empty string.
#config.environmentType = ""

# For CMS jobs, this is the CMSSW work area src directory where `cmsenv` should be executed.
# This directory must exist already.
# This parameter is only relevant and mandatory when environmentType = 'cms'.
# Defaults to an empty string.
#config.cmsswDir = ""

# The input files, apart from input data in the storage, needed by the jobs.
# Do NOT include archive files (tarballs) in this list.
# A tarball (input sandbox) will be created containing all the files in this list.
# The tarball will be copied to, and unpacked in, each job's local scratch directory on the worker nodes.
# Do NOT use absolute paths. Keep in mind also that relative paths are not only to locate the files locally,
# but also when the tarball is unpacked the relative paths are respected. That is, if you add in the input sandbox a line "mydir/*.txt",
# then all txt files present in ./mydir will be included in the input sandbox and when unpacked on the worker nodes
# you will have a directory mydir with all these txt files inside.
# Defaults to an empty list (i.e. no input sandbox).
#config.inputSandboxContent = []

# A directory (in a scratch or user area) where the input sandbox should be placed.
# If the directory doesn't exist, it will be created by the slurm_submit command.
# The jobs will copy the input sandbox from this directory once they start running.
# Defaults to the directory defined in 'sbatch_chdir'.
#config.inputSandboxDir = config.sbatch_chdir

# The filename of the input sandbox tarball.
# It is recommended to use the default in order to avoid overwritting by mistake other input sandboxes.
# Defaults to "input_sandbox_<timestamp>_<4-random-characters>.tar.gz"
#config.inputSandboxFilename = ""

# A directory where the SLURM batch script(s) should be written.
# If the directory doesn't exist, it will be created by the slurm_submit command.
# Defaults to the directory defined in 'sbatch_chdir'.
#config.batchScriptsDir = config.sbatch_chdir

# The filename of the SLURM batch script(s).
# In case of job arrays, only one batch script is created with the filename given by this parameter.
# In case of independent jobs, one batch script per job is created. The filenames are constructed from this parameter
# by appending "_<job-number>" just before the filename extension if there is one, or at the end otherwise.
# It is recommended to use the default in order to avoid overwritting by mistake other batch scripts.
# Defaults to "slurm_batch_script_<timestamp>_<4-random-characters>.sh" for job arrays,
# "slurm_batch_script_<timestamp>_<4-random-characters>_<job-number>.sh" otherwise.
#config.batchScriptsFilename = ""

# This flag tells whether to copy user files back from the worker node. 
# The files that will be copied back are the ones listed in 'stageoutFiles'.
# The stageout location should be specified in 'stageoutDir'.
# The copy is done by the batch script, after the payload.
# Note: the stageout of the job's stdout/stderr log file(s) is controlled by a separate flag ('stageoutLogs').
# Defaults to True.
#config.stageout = True

# A list with all the user files you want to stage out from the worker node.
# Only needed if 'stageout' is True.
# You don't need to include the job's stdout/stderr log file(s) in this list.
# The batch script will use the 'cp' command on each element of the list (so you can use wildcards in the filenames).
# Defaults to an empty list (i.e. no user files are copied back).
#config.stageoutFiles = []

# The stageout directory (in a scratch or user area) for the user files.
# If the directory doesn't exist, it will be created by the slurm job.
# Only needed if 'stageout' is True.
# Defaults to the directory defined in 'sbatch_chdir'.
#config.stageoutDir = config.sbatch_chdir

# When this flag is True (the default), the job's stdout/stderr log file(s) will be written
# locally on the worker nodes and the configuration parameters 'sbatch_output'
# and 'sbatch_error' will be both set to "/dev/null".
# To retrieve the log file(s) at the end of the job execution, set 'stageoutLogs' to True (the default).
# Defaults to True.
#config.writeLogsOnWN = True

# This flag is only relevant when 'writeLogsOnWN' is True.
# By default, stdout and stderr are redirected to the same file. If this flag is set to True,
# they will instead be redirected to different files.
# Defaults to False.
#config.separateStdoutStderrLogs = False

# This parameter is only relevant when 'writeLogsOnWN' is True.
# If 'separateStdoutStderrLogs' is False (the default), this parameter sets the name of the stdout/stderr log file.
# If 'separateStdoutStderrLogs' is True, this parameter sets the name of the stdout log file.
# Only file basenames are allowed; i.e. no directory paths (not even relative paths).
# The file will be created in each job's local scratch directory on the worker nodes.
# Defaults to "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out" for job arrays, "slurm-${SLURM_JOB_ID}.out" otherwise.
#config.stdoutFilename = "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.out"

# This parameter is only relevant when 'writeLogsOnWN' and 'separateStdoutStderrLogs' are both True,
# in which case this parameter sets the name of the stderr log file.
# Only file basenames are allowed; i.e. no directory paths (not even relative paths).
# The file will be created in each job's local scratch directory on the worker nodes.
# Defaults to "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err" for job arrays, "slurm-${SLURM_JOB_ID}.err" otherwise.
#config.stderrFilename = "slurm-${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}.err"

# This flag tells whether to copy the job's stdout/stderr log file(s) back from the worker node.
# This flag is only relevant when 'writeLogsOnWN' is True.
# The stageout location should be specified in 'stageoutLogsDir'.
# The copy is done by the batch script, at the very end (just before exiting).
# You don't need to include the log filename in the 'stageoutFiles' list.
# Defaults to True.
#config.stageoutLogs = True

# The stageout directory (in a scratch or user area) for the job's stdout/stderr log file(s).
# If the directory doesn't exist, it will be created by the slurm job.
# Only needed if 'stageoutLogs' is True.
# Defaults to a subdirectory named "logs" inside the directory defined in 'stageoutDir'.
#config.stageoutLogsDir = config.stageoutDir + "/logs"

# This flag tells whether to work with a SLURM job array or with indenpendent jobs.
# Defaults to True.
#config.useJobArray = True

# This parameter is only used when 'useJobArray' is True.
# It must be a non-negative integer (or None).
# It specifies the maximum number of jobs from the job array that can be simultaneously running.
# A value of 0 or None means no limit.
# Defaults to None.
#config.maxRunningJobs = None

# This parameter must be a positive integer. It specifies how many jobs to submit.
# If you work with a job array, the array will contain (the first) `numJobs` jobs.
# If you work with independent jobs, the meaning is obvious.
# You can set it to None to deactivate it. In that case the number of jobs to submit
# will be determined by the list of input parameters.
# Defaults to None.
#config.numJobs = None

#--------------------------------------------------------------------------------
# 3 Job-specific input parameters and payload
#--------------------------------------------------------------------------------

# Job input parameters will be passed to the job via bash environment variables.
# There will be one variable per input parameter (e.g. if each of your jobs needs
# N input parameters, there will be N bash environment variables defined in each job).
# You will have to provide the variable names in the configuration parameter 'inputParamsNames'.
# The same variable names will be used for all jobs, while their values will be set per job.
# Then it is the user responsability to correctly use these variables in his/ber payload.
# The payload is a set of bash lines (which can execute programs in any other language of course).
# So one way to retrieve the input variables is to do it in the bash lines of the payload.
# The way to access a variable in bash is with the following syntax: ${variable-name}.

# This configuration parameter should be a list containing the names of the job input parameters.
# It is important that the names in this list and the values in the 'inputParams' list
# follow the same order (the slurm_submit tool will assume that this is the case).
# Defaults to an empty list (i.e. no input parameters).
#config.inputParamsNames = []

# This configuration parameter should be a list where each element is itself a list containing
# the values of the job input parameters.
# It is important that the values in this list and the names in the 'inputParamsNames' list
# follow the same order (the slurm_submit tool will assume that this is the case).
# Defaults to an empty list (i.e. no input parameters).
#config.inputParams = []

# The payload is code that you want the job to run. It consists of a set of bash lines.
# In principle, it can be as simple as "./my_code" (remember to put the file 'my_code' in the input sandbox)
# and inside 'my_code' you do whatever you want.
# For CMS jobs, the payload can be for example "cmsRun my_code".
# (Remember: for CMS jobs, the `cmsenv` command does not need to be included in the payload;
# if 'cmsswDir' is specified, `cmsenv` will be executed before running the payload inside that directory).
# No validation of the content of this parameter is done.
# If you want to propagate an error inside the payload to the job exit code, make sure the payload
# exits with the right exit code. Following the suggestion in http://www.tldp.org/LDP/abs/html/exitcodes.html
# user-defined exit codes should be in the range 79-113, but we reserved the range 100-113
# for batch script errors. Payload exit codes in the range 100-113 will be propagated
# as job exit code 103 (the general exit code used by the batch script for signaling error in user payload).
# This is a schematic example of how to use exit codes in a payload:
# config.payload = \
# """
# do something ...
# if `do something ...` failed
#    exit 79
# do something else ...
# if `do something else ...` failed
#     exit 80
# exit 0
# """
# This parameter is mandatory.
# Defaults to an empty string.
#config.payload = ""

#--------------------------------------------------------------------------------
# 4 Checkpointing
#
# So far only the signaling to the payload was implemented. The user should implement a chackpointing in the payload.
#--------------------------------------------------------------------------------

# This flag tells whether to enable checkpointing (only the signaling to the payload).
# To disable the checkpointing, you can set 'checkpointSignal' to an empty string
# and 'checkpointSecsBeforeTimeout' to 0.
# Defaults to False.
#config.doCheckpointing = False

# This parameter is only relevant when 'doCheckpointing' is True.
# For the moment only signals 'SIGUSR1' and 'SIGUSR2' are allowed.
# When receiving and trapping this signal, the job will kill the payload and exit.
# It is the job of the payload to do the checkpointing.
# In particular, if 'checkpointSecsBeforeTimeout' is greater than 0, this signal
# will be send to the batch shell 'checkpointSecsBeforeTimeout' seconds before the time allocation limit.
# Note that the scancel command also sends signals to jobs. So if you do scancel
# with the signal defined here, the job will be checkpointed and exit.
# A job that has trapped the signal and exited,
# is expected to have status FAILED with SLURM exit code 10 (for SIGUSR1) or 12 (for SIGUSR2).
# If set to an empty string, no system signal will be trapped for checkpoint-and-exit.
# Defaults to an empty string.
#config.checkpointSignal = ""

# This parameter is only relevant when 'doCheckpointing' is True.
# This parameter specifies how many seconds before the time allocation limit to send a system signal 'checkpointSignal' to the job.
# It must be an integer.
# If set to 0 (or less than 0), the signal will not be programmed.
# It uses the SLURM sbatch option --signal. So you can not define this parameter
# and at the same time use the SLURM sbatch option --signal for other purpose.
# Defaults to 0.
#config.checkpointSecsBeforeTimeout = 0




"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import collections
import glob
import os
import re
import signal
import subprocess

import CP3SlurmUtils.Py2Py3Helpers
from CP3SlurmUtils.Configuration import loadConfiguration
from CP3SlurmUtils.Configuration import Configuration
from CP3SlurmUtils.ConfigurationUtils import validateConfigBeforeDefaults
from CP3SlurmUtils.ConfigurationUtils import fixConfigDefaults
from CP3SlurmUtils.ConfigurationUtils import setConfigDefaults
from CP3SlurmUtils.ConfigurationUtils import validateConfigAfterDefaults
from CP3SlurmUtils.ConfigurationUtils import fixConfigDirs
from CP3SlurmUtils.ConfigurationUtils import createConfigDirs
from CP3SlurmUtils.Exceptions import CP3SlurmUtilsException
from CP3SlurmUtils.InteractiveQuestion import interactiveYesNoQuestion
from CP3SlurmUtils.InteractiveQuestion import questionTimeoutHandler
from CP3SlurmUtils.InteractiveQuestion import TimeoutException
from CP3SlurmUtils.SubmitUtils import submit


class SubmitWorker:

    def __init__(self, config=None, submit=False, yes=False, summaryJobs='', debug=False, quiet=False, calledFromScript=False):
        """SubmitWorker constructor."""

        self.calledFromScript = calledFromScript

        if isinstance(debug, bool):
            self.debug = debug
        else:
            msg = "ERROR: {} argument 'debug' must be a boolean.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        if isinstance(quiet, bool):
            self.quiet = quiet
        else:
            msg = "ERROR: {} argument 'quiet' must be a boolean.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        if isinstance(submit, bool):
            self.submit = submit
        else:
            msg = "ERROR: {} argument 'submit' must be a boolean.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        if isinstance(yes, bool):
            self.yes = yes
        else:
            msg = "ERROR: {} argument 'yes' must be a boolean.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        if isinstance(summaryJobs, str):
            self.summaryJobs = []
            for v in [j for j in map(str.strip, summaryJobs.split(',')) if j]:
                if not re.match('^[1-9][0-9]*$', v):
                    if self.calledFromScript:
                        msg = "ERROR: Option --summary-jobs has an invalid value: '{value}'. It must be a comma separated list of positive integers.".format(value=v)
                    else:
                        msg = "ERROR: Argument 'summaryJobs' has an invalid value: '{value}'. It must be a comma separated list of positive integers.".format(value=v)
                    raise CP3SlurmUtilsException(msg)
                if int(v) not in self.summaryJobs:
                    self.summaryJobs.append(int(v))
            self.summaryJobs = sorted(self.summaryJobs)
        else:
            msg = "ERROR: {} argument 'summaryJobs' must be a string.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        self.config = None
        self.configValidated = False
        if config is not None:
            self.loadConfig(config)


    def loadConfig(self, config):
        """Load and validate the configuration."""

        self.configValidated = False

        if isinstance(config, Configuration):
            self.config = config
        elif isinstance(config, str):
            if not os.path.isfile(config):
                msg = "ERROR: Could not find configuration file {file}".format(file=config)
                raise CP3SlurmUtilsException(msg)
            if not self.quiet:
                print("Using configuration file {file}".format(file=config))
            self.config = loadConfiguration(config)
            if self.debug:
                print("Configuration loaded.")
        else:
            msg = "ERROR: {} argument 'config' must be a Configuration object or a string representing a filename.".format(self.__class__.__name__)
            raise CP3SlurmUtilsException(msg)

        self.validateConfig()


    def validateConfig(self):
        """Validate the configuration."""

        validateConfigBeforeDefaults(self.config)

        if self.debug:
            print("Configuration passed first validation.")

        fixConfigDefaults(self.config)

        if self.debug:
            print("Configuration parameter defaults redefined.")

        setConfigDefaults(self.config)

        if self.debug:
            print("Configuration parameter defaults assigned.")

        validateConfigAfterDefaults(self.config)

        if self.debug:
            print("Configuration passed final validation.")

        fixConfigDirs(self.config)

        if self.debug:
            print("Directory paths in configuration completed.")

        createConfigDirs(self.config)

        if self.debug:
            print("Directories from configuration created.")

        self.configValidated = True


    def __call__(self):
        """Run all the workload for submission."""

        if self.config is None:
            msg = "ERROR: There is no configuration defined."
            raise CP3SlurmUtilsException(msg)

        if not self.quiet:
            print("Preparing the jobs ...")

        if not self.configValidated:
            self.validateConfig()

        # For commodity, put all the configuration parameters in local variables
        # and from now on the config object should not be used anymore.
        # If some parameters need to be adjusted, it should not be done on the config object, but on these variables.
        sbatch_partition            = self.config.sbatch_partition
        sbatch_qos                  = self.config.sbatch_qos
        sbatch_chdir                = self.config.sbatch_chdir
        sbatch_time                 = self.config.sbatch_time
        sbatch_memPerCPU            = self.config.sbatch_memPerCPU
        sbatch_output               = self.config.sbatch_output
        sbatch_error                = self.config.sbatch_error
        sbatch_additionalOptions    = self.config.sbatch_additionalOptions
        writeLogsOnWN               = self.config.writeLogsOnWN
        stdoutFilename              = self.config.stdoutFilename
        stderrFilename              = self.config.stderrFilename
        separateStdoutStderrLogs    = self.config.separateStdoutStderrLogs
        scratchDir                  = self.config.scratchDir
        handleScratch               = self.config.handleScratch
        environmentType             = self.config.environmentType
        cmsswDir                    = self.config.cmsswDir
        inputSandboxContent         = self.config.inputSandboxContent
        inputSandboxDir             = self.config.inputSandboxDir
        inputSandboxFilename        = self.config.inputSandboxFilename
        batchScriptsDir             = self.config.batchScriptsDir
        batchScriptsFilename        = self.config.batchScriptsFilename
        stageout                    = self.config.stageout
        stageoutDir                 = self.config.stageoutDir
        stageoutFiles               = self.config.stageoutFiles
        stageoutLogs                = self.config.stageoutLogs
        stageoutLogsDir             = self.config.stageoutLogsDir
        useJobArray                 = self.config.useJobArray
        maxRunningJobs              = self.config.maxRunningJobs
        numJobs                     = self.config.numJobs
        inputParamsNames            = self.config.inputParamsNames
        inputParams                 = self.config.inputParams
        doCheckpointing             = self.config.doCheckpointing
        checkpointSignal            = self.config.checkpointSignal
        checkpointSecsBeforeTimeout = self.config.checkpointSecsBeforeTimeout
        payload                     = self.config.payload

        if writeLogsOnWN:
            sbatch_output = '/dev/null'
            sbatch_error = '/dev/null'

        environmentType = environmentType.strip().lower()
        if self.debug:
            print("The environment type is '{env}'.".format(env=environmentType))

        numJobInputParams = len(inputParamsNames)
        if numJobs is None:
            numJobs = len(inputParams)
        elif numJobs < len(inputParams):
            if self.debug:
                msg = "The 'inputParams' list in the configuration contains more parameters than necessary for {num} jobs."
                msg += " Will use only the first {num} sets of input parameters from 'inputParams'."
                msg = msg.format(num=numJobs)
                print(msg)
            inputParams = inputParams[:numJobs]

        if self.summaryJobs and numJobInputParams == 0:
            if not self.quiet:
                if self.calledFromScript:
                    print("Ignoring option --summary-jobs, because no input parameters for the job's payload were specified" + \
                          " (the job's payload input parameters are the only information that would be shown in the summary at a per job basis).")
                else:
                    print("Ignoring argument 'summaryJobs', because no input parameters for the job's payload were specified" + \
                          " (the job's payload input parameters are the only information that would be shown in the summary at a per job basis).")
            self.summaryJobs = []
        for i, v in enumerate(self.summaryJobs):
            if v > numJobs:
                if not self.quiet:
                    if self.calledFromScript:
                        msg = "Option --summary-jobs has values that are bigger than the total number of jobs ({num}): {0}."
                        msg += " Will ignore these particular values."
                        msg = msg.format(self.summaryJobs[i:], num=numJobs)
                        print(msg)
                    else:
                        msg = "Argument 'summaryJobs' has values that are bigger than the total number of jobs ({num}): {0}."
                        msg += " Will ignore these particular values."
                        msg = msg.format(self.summaryJobs[i:], num=numJobs)
                        print(msg)
                self.summaryJobs = self.summaryJobs[:i]
                break

        jobArrayTaskIds = ''
        if useJobArray:
            if maxRunningJobs is None:
                maxRunningJobs = 0
            if self.debug:
                print("Will prepare a batch script for a slurm job array of {num} jobs.".format(num=numJobs))
                if maxRunningJobs > 0:
                    print("Will limit the number of concurrently running jobs to {}.".format(maxRunningJobs))
            if numJobs == 1:
                jobArrayTaskIds = '1'
            else:
                jobArrayTaskIds = '1-{}'.format(numJobs)
                if maxRunningJobs > 0:
                    jobArrayTaskIds += '%{}'.format(maxRunningJobs)
            if self.debug:
                print("Defined job array task ids as '{ids}'.".format(ids=jobArrayTaskIds))
        else:
            if self.debug:
                print("Will prepare {num} batch scripts for {num} independent jobs.".format(num=numJobs))

        #================================================================================
        # Define lists with the job log files and with the stageout files
        #================================================================================

        logFiles = []
        if writeLogsOnWN:
            if stageoutLogs:
                logFilesWN = [stdoutFilename]
                if separateStdoutStderrLogs:
                    logFilesWN += [stderrFilename]
                for f in logFilesWN:
                    logFiles.append('{}/{}'.format(stageoutLogsDir, f))
        else:
            if sbatch_output:
                if sbatch_output != '/dev/null':
                    if os.path.isabs(sbatch_output):
                        logFiles.append(sbatch_output)
                    else:
                        logFiles.append('{}/{}'.format(sbatch_chdir, sbatch_output))
            else:
                if useJobArray:
                    logFiles.append('{}/slurm-${{SLURM_ARRAY_JOB_ID}}_${{SLURM_ARRAY_TASK_ID}}.out'.format(sbatch_chdir))
                else:
                    logFiles.append('{}/slurm-${{SLURM_JOB_ID}}.out'.format(sbatch_chdir))
            if sbatch_error:
                if sbatch_error != '/dev/null':
                    if os.path.isabs(sbatch_error):
                        logFiles.append(sbatch_error)
                    else:
                        logFiles.append('{}/{}'.format(sbatch_chdir, sbatch_error))

        stageoutFiles = [s.strip() for s in stageoutFiles if s.strip()]

        #================================================================================
        # Prepare the input sandbox
        #================================================================================

        inputSandbox = ''
        if inputSandboxContent:
            if self.debug:
                print("Creating input sandbox.")
            inputSandbox = '{}/{}'.format(inputSandboxDir, inputSandboxFilename)
            args = ['tar', '-czf', inputSandbox]
            missingFiles = []
            for isc in inputSandboxContent:
                if os.path.isabs(isc):
                    msg = "ERROR: The files to add to the input sandbox should be given with relative paths, not absolute paths."
                    if self.calledFromScript:
                        msg += " The paths should be relative to the directory from where you are running the slurm_submit command."
                    raise CP3SlurmUtilsException(msg)
                files = glob.glob(isc)
                if not files:
                    missingFiles.append(isc)
                else:
                    args += files
            if missingFiles:
                msg = "ERROR: Some of the files that were requested to be added to the input sandbox were not found: {files}.".format(files=missingFiles)
                raise CP3SlurmUtilsException(msg)
            else:
                returncode = subprocess.call(args)
            if returncode != 0:
                msg = "ERROR: Failed to create input sandbox."
                raise CP3SlurmUtilsException(msg)
            if self.debug:
                print("Input sandbox created.")
        else:
            if self.debug:
                print("No input sanbox was created.")

        #================================================================================
        # Prepare the bash code that will set the job input parameters
        #================================================================================

        # A few lines of bash code to put the job input parameters in environment variables.
        # This code will be inserted before the payload.
        # For independent jobs we have one set of lines of code per job.
        setInputParamsCode = []
        if numJobInputParams > 0:
            if self.debug:
                print("Preparing bash code for reading input parameters for job's payload.")
            if useJobArray:
                inputParamsStr = ""
                for j in range(numJobs):
                    inputParamsStr += "\n\"{}\"".format(' <sep> '.join(list(map(str, inputParams[j]))))
                    if j == numJobs - 1: inputParamsStr += "\n"
                setInputParamsCodeStr = "inputParams=({})".format(inputParamsStr)
                setInputParamsCodeStr += "\njobInputParams=\"${inputParams[${SLURM_ARRAY_TASK_ID}-1]}\""
                for i in range(numJobInputParams):
                    setInputParamsCodeStr += "\nexport {paramName}=`echo -e \"${{jobInputParams}}\" | awk -F' <sep> ' '{{print ${paramNum}}}'`".format(paramName=inputParamsNames[i], paramNum=i+1)
                    setInputParamsCodeStr += "\necho \"{paramName} = ${{{paramName}}}\"".format(paramName=inputParamsNames[i])
                setInputParamsCode.append(setInputParamsCodeStr)
            else:
                for j in range(numJobs):
                    setInputParamsCodeStr = ""
                    for i in range(numJobInputParams):
                        if i > 0:
                            setInputParamsCodeStr += "\n"
                        setInputParamsCodeStr += "export {paramName}={paramValue}".format(paramName=inputParamsNames[i], paramValue=inputParams[j][i])
                        setInputParamsCodeStr += "\necho \"{paramName} = ${{{paramName}}}\"".format(paramName=inputParamsNames[i])
                    setInputParamsCode.append(setInputParamsCodeStr)
            if self.debug:
                print("Code for reading input parameters for job's payload created.")
        else:
            setInputParamsCodeStr = "echo \"No input parameters were defined for the payload.\""
            if self.debug:
                print("No code for reading input parameters for job's payload was created.")

        #================================================================================
        # Construct the batch script template
        #================================================================================

        if self.debug:
            print("Constructing batch scripts template.")

        batchScriptTemplate = \
"""\
#!/bin/bash
#
#SBATCH --chdir=__SBATCH_CHDIR__
#SBATCH --time=__SBATCH_TIME__
#SBATCH --mem-per-cpu=__SBATCH_MEM_PER_CPU__
#SBATCH --partition=__SBATCH_PARTITION__
"""
        if sbatch_qos:
            batchScriptTemplate += \
"""\
#SBATCH --qos=__SBATCH_QOS__
"""
        if useJobArray:
            batchScriptTemplate += \
"""\
#SBATCH --array=__SBATCH_ARRAY__
"""
        if sbatch_output:
            batchScriptTemplate += \
"""\
#SBATCH --output=__SBATCH_OUTPUT__
"""
        if sbatch_error:
            batchScriptTemplate += \
"""\
#SBATCH --error=__SBATCH_ERROR__
"""
        if doCheckpointing and checkpointSecsBeforeTimeout > 0:
            batchScriptTemplate += \
"""\
#SBATCH --signal=B:__CHECKPOINT_SIGNAL_SHORT__@__CHECKPOINT_SECS_BEFORE_TIMEOUT__
"""
        if sbatch_additionalOptions:
            batchScriptTemplate += \
"""\
__SBATCH_ADDITIONAL_OPTIONS__
"""
        sbatchAdditionalOptionsStr = ""
        if sbatch_additionalOptions:
            for i, sbatchAddOpt in enumerate(sbatch_additionalOptions):
                sbatchAdditionalOptionsStr += "#SBATCH {}".format(sbatchAddOpt)
                if i < len(sbatch_additionalOptions)-1:
                    sbatchAdditionalOptionsStr += "\n"
        batchScriptTemplate += \
"""\

#==================================================================================

# This variable keeps the exit code with which the batch script will exit.
exitCode=0
exitCodeMeaning=""

function updateExitCode() {
    [[ ${exitCode} -ne 0 ]] || exitCode=${1}
}

startupWorkDir=`pwd`
"""
        if scratchDir:
            if handleScratch:
                batchScriptTemplate += \
"""\

#==================================================================================
# Create the job's scratch directory
#==================================================================================

if [[ ! -d __SCRATCH_DIR__ ]]
then
    mkdir -p __SCRATCH_DIR__
fi
"""
            batchScriptTemplate += \
"""\

#==================================================================================
# Change directory to the job's scratch directory
#==================================================================================

scratchDirErrorMsg=""
if [[ -d __SCRATCH_DIR__ ]]
then
    cd __SCRATCH_DIR__ || updateExitCode 100
else
    scratchDirErrorMsg="Local scratch directory __SCRATCH_DIR__ does not exist."
    updateExitCode 100
fi
"""
        batchScriptTemplate += \
"""\

#==================================================================================
# Redirect stdout and stderr to local file(s)
#==================================================================================

writeLogsOnWN=__WRITE_LOGS_ON_WN__
separateStdoutStderrLogs=__SEPARATE_STDOUT_STDERR_LOGS__
stdoutFilename="__STDOUT_FILENAME__"
stderrFilename="__STDERR_FILENAME__"

if [[ "${writeLogsOnWN}" == "true" ]]
then
    # Redirect stdout and stderr to local file(s).
    if [[ "${separateStdoutStderrLogs}" == "true" ]]
    then
        exec 1> "${stdoutFilename}" 2> "${stderrFilename}"
    else
        exec 1> "${stdoutFilename}" 2>&1
    fi
fi

#==================================================================================
# Print some info about the job
#==================================================================================

echo "======== Starting job (`date`) ========"
echo "Job id: ${SLURM_JOB_ID}"
echo "List of nodes allocated to the job: ${SLURM_JOB_NODELIST}"
echo "Batch node: ${SLURMD_NODENAME}"
"""
        if scratchDir:
            batchScriptTemplate += \
"""\
echo "Startup working directory: ${startupWorkDir}"

if [[ ${exitCode} -eq 0 ]]
then
    echo "Job working directory: `pwd`"
elif [[ "${scratchDirErrorMsg}" != "" ]]
then
    echo ${scratchDirErrorMsg}
fi
"""
        else:
            batchScriptTemplate += \
"""\
echo "Working directory: ${startupWorkDir}"
"""
        if environmentType == 'cms':
            batchScriptTemplate += \
"""\

#==================================================================================
# Set up the environment
#==================================================================================

cmsswDir="__CMSSW_DIR__"

function setupEnvironment() {
    # Set up the CMS environment
    echo "Sourcing CMS environment file /cvmfs/cms.cern.ch/cmsset_default.sh"
    source /cvmfs/cms.cern.ch/cmsset_default.sh
    [[ $? -eq 0 ]] || return 1
    # Set up the runtime CMS environment for the given CMSSW work area
    if [[ ! -z "${cmsswDir}" ]]
    then
        echo "CMSSW work area directory: ${cmsswDir}"
        if [[ ! -d "${cmsswDir}" ]]
        then
            echo "CMSSW work area directory does not exist."
            return 1
        fi
        currentDir=`pwd`
        cd "${cmsswDir}" || return 1
        echo "Running cmsenv"
        eval `scramv1 runtime -sh`
        cd "${currentDir}"
    fi
}

if [[ ${exitCode} -eq 0 ]]
then
    echo "==== Starting setup of CMS environment (`date`) ===="
    setupEnvironment || updateExitCode 101
    echo "==== Finished setup of CMS environment (`date`) ===="
fi
"""
        if inputSandbox:
            batchScriptTemplate += \
"""\

#==================================================================================
# Stage in and unpack the input sandbox
#==================================================================================

inputSandbox="__INPUT_SANDBOX__"
inputSandboxDir="$(dirname ${inputSandbox})"

function copy_and_unpack_input_sandbox() {
    case "${inputSandbox}" in
        *\ * )
            echo "Only one input sandbox (max. size 100 MB) should be copied. Will not continue."
            return 1
            ;;
    esac
    echo "Input sandbox file: ${inputSandbox}"
    if [[ ! -f "${inputSandbox}" ]]
    then
        echo "Input sandbox does not exist."
        return 1
    fi
    # Check that input sandbox is not bigger than 100 MB.
    local inputSandboxSize=`du -m "${inputSandbox}" | awk '{print $1}'`
    if [[ ${inputSandboxSize} -gt 100 ]]
    then
        echo "Input sandbox size (${inputSandboxSize} MB) greater than 100 MB. Will not continue."
        return 1
    fi
    echo "Input sanbox contents:"
    tar -ztvf "${inputSandbox}" | awk '{$1=$2=$3=$4=$5=""; print $0}'
    if [[ ${SLURM_JOB_NUM_NODES} -gt 1 ]]
    then
        execFile="${inputSandboxDir}/stagein-ISB"
        echo "Will create executable file ${execFile} to be run via srun in order to stagein the input sandbox to all allocated worker nodes." 
        if [[ -f "${execFile}" ]]
        then
            echo "A file ${execFile} already exists. Please remove it or rename it."
            return 1
        fi
        cat > "${execFile}" <<EOF
#!/bin/bash

function run_cmd() {
    cmd=("\${@}")
    ec=0
"""
            if scratchDir:
                batchScriptTemplate += \
"""\
    tmplog=__SCRATCH_DIR__/$(basename ${execFile}).err
"""
            else:
                batchScriptTemplate += \
"""\
    tmplog=/tmp/$(basename ${execFile}).err
"""
            batchScriptTemplate += \
"""\
    "\${cmd[@]}" 2> \${tmplog} >/dev/null || ec=1
    while read line
    do
        >&2 echo "\`hostname -f\`: \$line"
    done < \${tmplog};
    rm \${tmplog}
    return \$ec
}

inputSandbox="\${1}"
echo "\`hostname -f\`: Copying and unpacking input sandbox ..."
run_cmd cp "\${inputSandbox}" .
if [[ \$? -eq 0 ]]
then
    run_cmd tar -zxf "\$(basename \${inputSandbox})"
    if [[ \$? -eq 0 ]]
    then
        echo "\`hostname -f\`: Input sandbox copied and unpacked."
    else
        echo "\`hostname -f\`: Failed to unpack input sandbox."
        exit 1
    fi
else
    echo "\`hostname -f\`: Failed to copy input sandbox."
    exit 1
fi
EOF
        chmod 0700 "${execFile}"
        if [[ -x "${execFile}" ]]
        then
            echo "Executable file ${execFile} created."
        else
            echo "Failed to create executable file ${execFile}"
            return 1
        fi
        srun --ntasks-per-node=1 "${execFile}" "${inputSandbox}"
        ec=$?
        echo "Removing previously created file ${execFile}"
        rm -f "${execFile}"
        return $ec
    else
        echo "Copying and unpacking input sandbox ..."
        cp "${inputSandbox}" .
        if [[ $? -eq 0 ]]
        then
            tar -zxf "$(basename ${inputSandbox})"
            if [[ $? -eq 0 ]]
            then
                echo "Input sandbox copied and unpacked."
            else
                echo "Failed to unpack input sandbox."
                return 1
            fi
        else
            echo "Failed to copy input sandbox."
            return 1
        fi
    fi
}

if [[ ${exitCode} -eq 0 ]]
then
    echo "==== Starting stagein of input sandbox (`date`) ===="
    if [[ -n "${inputSandbox}" ]]
    then
        copy_and_unpack_input_sandbox || updateExitCode 102
    else
        echo "No input sandbox was specified for stagein."
    fi
    echo "==== Finished stagein of input sandbox (`date`) ===="
fi
"""
        batchScriptTemplate += \
"""\

#==================================================================================
# Read the job input arguments (if any) and run the payload
#==================================================================================
"""
        if doCheckpointing and checkpointSignal:
            batchScriptTemplate += \
"""\

function checkpoint() {
    echo "Checkpointing is left up to the user ..."
}

function kill_payload() {
    echo "Killing payload with signal __CHECKPOINT_SIGNAL__."
    kill -s __CHECKPOINT_SIGNAL__ ${payloadpid}
    wait ${payloadpid}
    pec=$?
}

function signalhandler() {
    echo "Received signal __CHECKPOINT_SIGNAL__."
    checkpoint
    kill_payload
}
"""
        if setInputParamsCode:
            batchScriptTemplate += \
"""\

if [[ ${exitCode} -eq 0 ]]
then
    echo "==== Starting read of input parameters for job's payload (`date`) ===="
    __SET_INPUT_PARAMS_CODE__
    echo "==== Finished read of input parameters for job's payload (`date`) ===="
fi
"""
        if doCheckpointing and checkpointSignal:
            batchScriptTemplate += \
"""\
trap "signalhandler" __CHECKPOINT_SIGNAL__
"""
        batchScriptTemplate += \
"""\

if [[ ${exitCode} -eq 0 ]]
then
    echo "==== Starting execution of payload (`date`) ===="
    echo "------------------------- Begin payload output ------------------------"
    (
__PAYLOAD__
    ) &
    payloadpid=$!
    wait ${payloadpid}
    pec=${pec:-$?}
    echo "--------------------------- End payload output ------------------------"
    echo "Payload exit code: ${pec}"
    if [[ ${pec} -ne 0 ]]
    then
        if [[ ${pec} -ge 100 && ${pec} -le 113 ]]
        then
            echo "Payload exit code is in range 100-113, which is reserved for batch script non-payload errors." \\
                 "Setting batch script exit code to 103."
            updateExitCode 103
        else
            if [[ ${pec} -lt 79 || ${pec} -gt 99 ]]
            then
                echo "Payload exit code is outside the recommended range for user-defined exit codes: 79-99." \\
                     "This is not a problem, but if you are interested in defining/using exit codes for your payload that would not occur outside your payload," \\
                     "please consider restricting to that range."
            fi
            updateExitCode ${pec}
            exitCodeMeaning="failure in user's payload"
        fi
    fi
    echo "==== Finished execution of payload (`date`) ===="
fi

#==================================================================================
# Stageout of user files
#==================================================================================

function stageout() {
    local filetype="${1}"
    local files=("")
    local dir=""
    if [[ "${filetype}" == "userfiles" ]]
    then
        files=("${stageoutFiles[@]}")
        dir="${stageoutDir}"
    elif [[ "${filetype}" == "logs" ]]
    then
        files=("${logFiles[@]}")
        dir="${stageoutLogsDir}"
    fi
    if [[ ! -n "${files}" ]]
    then
        echo "No files were specified for stageout."
        return 0
    fi
    if [[ ! -n "${dir}" ]]
    then
        echo "WARNING: The following files were requested for stageout:"
        echo "${files[@]}"
        echo "but no stageout directory was specified where to copy the files."
        echo "Will NOT stage out the files."
        return 0
    fi
    mkdir -p "${dir}"
    if [[ "${filetype}" == "logs" ]]
    then
        cp "${files[@]}" "${dir}/"
        return 0
    fi
    echo "Destination directory: ${dir}"
    if [[ ! -d "${dir}" ]]
    then
        echo "Destination directory does not exist and failed to create it."
        return 1
    fi
    function join_by() { local d=$1; shift; echo -n "$1"; shift; printf "%s" "${@/#/$d}"; }
    patterns=$(join_by "', '" "${files[@]}")
    echo "List of filename patterns to match against: '${patterns}'."
    if [[ ${SLURM_JOB_NUM_NODES} -gt 1 ]]
    then
        execFile="${dir}/stageout-OSB"
        echo "Will create executable file ${execFile} to be run via srun in order to stageout the user files from all allocated worker nodes." 
        if [[ -f "${execFile}" ]]
        then
            echo "A file ${execFile} already exists. Please remove it or rename it."
            return 1
        fi
        cat > "${execFile}" <<EOF
#!/bin/bash

function run_cmd() {
    cmd=("\${@}")
    ec=0
"""
        if scratchDir:
            batchScriptTemplate += \
"""\
    tmplog=__SCRATCH_DIR__/$(basename ${execFile}).err
"""
        else:
            batchScriptTemplate += \
"""\
    tmplog=/tmp/$(basename ${execFile}).err
"""
        batchScriptTemplate += \
"""\
    "\${cmd[@]}" 2> \${tmplog} >/dev/null || ec=1
    while read line
    do
        >&2 echo "\`hostname -f\`: \$line"
    done < \${tmplog};
    rm \${tmplog}
    return \$ec
}

files=("\${@}")
for pattern in "\${files[@]}"
do
    matchingFiles=()
    while IFS=  read -r -d $'\\\\0'
    do
        matchingFiles+=("\$REPLY")
    done < <(find -mindepth 1 -maxdepth 1 -name "\${pattern}" -print0)
    if [[ \${#matchingFiles[@]} -gt 0 ]]
    then
        for file in "\${matchingFiles[@]}"
        do
            echo "\`hostname -f\`: Copying file \${file/\.\//}"
            run_cmd cp -r "\${file}" "${dir}/"
            if [[ \$? -eq 0 ]]
            then
                echo "\`hostname -f\`: File copied."
            else
                echo "\`hostname -f\`: Failed to copy file."
                exit 1
            fi
        done
    else
        echo "\`hostname -f\`: No files matching pattern '\${pattern}'."
        exit 1
    fi
done
EOF
        chmod 0700 "${execFile}"
        if [[ -x "${execFile}" ]]
        then
            echo "Executable file ${execFile} created."
        else
            echo "Failed to create executable file ${execFile}"
            return 1
        fi
        srun --ntasks-per-node=1 "${execFile}" "${files[@]}"
        ec=$?
        echo "Removing previously created file ${execFile}"
        rm -f "${execFile}"
        return $ec
    else
        for pattern in "${files[@]}"
        do
            matchingFiles=()
            while IFS=  read -r -d $'\\0'
            do
                matchingFiles+=("$REPLY")
            done < <(find -mindepth 1 -maxdepth 1 -name "${pattern}" -print0)
            if [[ ${#matchingFiles[@]} -gt 0 ]]
            then
                for file in "${matchingFiles[@]}"
                do
                    echo "Copying file ${file/\.\//} ..."
                    cp -r "${file}" "${dir}/"
                    if [[ $? -eq 0 ]]
                    then
                        echo "File copied."
                    else
                        echo "Failed to copy file."
                        return 1
                    fi
                done
            else
                echo "No files matching pattern '${pattern}'."
                return 1
            fi
        done
    fi
}

stageout=__STAGEOUT__
stageoutDir="__STAGEOUT_DIR__"
stageoutFiles=("__STAGEOUT_FILES__")

if [[ ${exitCode} -eq 0 ]]
then
    echo "==== Starting stageout of user files (`date`) ===="
    if [[ "${stageout}" == "true" ]]
    then
        stageout userfiles || updateExitCode 104
    else
        echo "Stageout flag is off. Will not stage out any user file."
    fi
    echo "==== Finished stageout of user files (`date`) ===="
fi

#==================================================================================
# Print a final exit status message
#==================================================================================

echo "======== Finished job (`date`) ========"

exitMsg="Batch script exit code: ${exitCode}"
if [[ ${exitCode} -eq 100 ]]
then
    exitCodeMeaning="failure in changing directory to the job's local scratch directory"
elif [[ ${exitCode} -eq 101 ]]
then
    exitCodeMeaning="failure in environment setup"
elif [[ ${exitCode} -eq 102 ]]
then
    exitCodeMeaning="failure in stagein/unpack of input sandbox"
elif [[ ${exitCode} -eq 103 ]]
then
    exitCodeMeaning="failure in user's payload"
elif [[ ${exitCode} -eq 104 ]]
then
    exitCodeMeaning="failure in stageout of user files"
fi
if [[ -n "${exitCodeMeaning}" ]]
then
    exitMsg+=" (${exitCodeMeaning})"
fi
echo "${exitMsg}"

#==================================================================================
# Stageout the logs and exit
#==================================================================================

stageoutLogs=__STAGEOUT_LOGS__
stageoutLogsDir="__STAGEOUT_LOGS_DIR__"

if [[ "${writeLogsOnWN}" == "true" && "${stageoutLogs}" == "true" ]]
then
    echo "Will stageout the logs before exiting."
    if [[ "${separateStdoutStderrLogs}" == "true" ]]
    then
        logFiles=("${stdoutFilename}" "${stderrFilename}")
    else
        logFiles=("${stdoutFilename}")
    fi
    stageout logs
fi
"""
        if scratchDir and handleScratch:
            batchScriptTemplate += \
"""\

#==================================================================================
# Remove the job's scratch directory
#==================================================================================

if [[ -d __SCRATCH_DIR__ ]]
then
    rm -rf __SCRATCH_DIR__
fi
"""
        batchScriptTemplate += \
"""\
#==================================================================================

exit ${exitCode}
"""

        if self.debug:
            print("Batch scripts template constructed.")

        #================================================================================
        # Create the bash script(s) from the template
        #================================================================================

        if self.debug:
            print("Creating batch script(s).")

        batchScripts = []

        replaceDict = {
            '__SBATCH_PARTITION__': sbatch_partition,
            '__SBATCH_QOS__': sbatch_qos,
            '__SBATCH_CHDIR__': sbatch_chdir,
            '__SBATCH_TIME__': sbatch_time,
            '__SBATCH_MEM_PER_CPU__': sbatch_memPerCPU,
            '__SBATCH_ARRAY__': jobArrayTaskIds,
            '__SBATCH_OUTPUT__': sbatch_output,
            '__SBATCH_ERROR__': sbatch_error,
            '__SBATCH_ADDITIONAL_OPTIONS__': sbatchAdditionalOptionsStr,
            '__WRITE_LOGS_ON_WN__': str(writeLogsOnWN).lower(),
            '__SEPARATE_STDOUT_STDERR_LOGS__': str(separateStdoutStderrLogs).lower(),
            '__STDOUT_FILENAME__': stdoutFilename,
            '__STDERR_FILENAME__': stderrFilename,
            '__SCRATCH_DIR__': scratchDir,
            '__CMSSW_DIR__': cmsswDir,
            '__INPUT_SANDBOX__': inputSandbox,
            '__PAYLOAD__': payload.strip('\n'),
            '__STAGEOUT__': str(stageout).lower(),
            '__STAGEOUT_DIR__': stageoutDir,
            '__STAGEOUT_FILES__': '" "'.join(stageoutFiles),
            '__STAGEOUT_LOGS__': str(stageoutLogs).lower(),
            '__STAGEOUT_LOGS_DIR__': stageoutLogsDir,
            '__CHECKPOINT_SECS_BEFORE_TIMEOUT__': str(checkpointSecsBeforeTimeout),
            '__CHECKPOINT_SIGNAL__': checkpointSignal,
            '__CHECKPOINT_SIGNAL_SHORT__': checkpointSignal.replace('SIG', ''),

        }

        for j in range(numJobs):
            if setInputParamsCode:
                replaceDict['__SET_INPUT_PARAMS_CODE__'] = setInputParamsCode[j].replace('\n', '\n    ')
            batchScriptText = batchScriptTemplate
            for k, v in CP3SlurmUtils.Py2Py3Helpers.iteritems(replaceDict):
                batchScriptText = batchScriptText.replace(k, v)
            if useJobArray:
                batchScript = '{}/{}'.format(batchScriptsDir, batchScriptsFilename)
            else:
                if '.' in batchScriptsFilename:
                    batchScript = '{}/{}_{}.{}'.format(batchScriptsDir, '.'.join(batchScriptsFilename.split('.')[:-1]), j+1, batchScriptsFilename.split('.')[-1])
                else:
                    batchScript = '{}/{}_{}'.format(batchScriptsDir, batchScriptsFilename, j+1)
            try:
                with open(batchScript, 'w') as fd:
                    fd.write(batchScriptText)
            except IOError as ex:
                msg = "ERROR: Failed to create batch script(s)."
                msg += "\nError follows:\n{}"
                msg = msg.format(str(ex))
                raise CP3SlurmUtilsException(msg)
            batchScripts.append(batchScript)
            if useJobArray:
                break

        if self.debug:
            print("Batch script(s) created.")

        #================================================================================
        # Print a summary of what will be submitted
        #================================================================================

        if self.debug:
            print("Creating summary text.")

        summary = "================================================================================"
        summary += "\n=============  Below there is a summary of what would be submitted  ============"
        summary += "\n================================================================================"
        if useJobArray:
            summary += "\nWould submit a job array consisting of {num} jobs.".format(num=numJobs)
        else:
            summary += "\nWould submit {num} independent jobs.".format(num=numJobs)
        if numJobInputParams > 0:
            if self.summaryJobs:
                if self.calledFromScript:
                    summary += "\nInput parameters to the job's payload for the jobs specified in the --summary-jobs option."
                else:
                    summary += "\nInput parameters to the job's payload for the jobs specified in the 'summaryJobs' argument."
            else:
                self.summaryJobs = list(range(1,min(10,numJobs)+1))
                summary += "\nInput parameters to the job's payload for the first {num} jobs:".format(num=min(10,numJobs))
            for j in self.summaryJobs:
                summary += "\nJob {}:".format(j)
                for i in range(numJobInputParams):
                    summary += "\n    {paramName} = {paramValue}".format(paramName=inputParamsNames[i], paramValue=inputParams[j-1][i])
        else:
            summary += "\nNo input parameters would be passed to the job's payload."
        if batchScripts:
            summary += "\nBatch scripts:"
            summary += "\n    {}".format(batchScripts[0])
            if not useJobArray and numJobs > 1:
                if numJobs > 2:
                    summary += "\n    ..."
                summary += "\n    {}".format(batchScripts[-1])
        if inputSandbox:
            summary += "\nInput sandbox:"
            summary += "\n    {}".format(inputSandbox)
        else:
            summary += "\nNo input sandbox would be staged in to the worker nodes."
        if stageout and stageoutFiles:
            summary += "\nStageout directory for user files:"
            summary += "\n    {}".format(stageoutDir)
        else:
            summary += "\nNo user files would be staged out from the worker nodes."
        if writeLogsOnWN:
            if stageoutLogs and stageoutLogsDir:
                summary += "\nJob's stdout/stderr log files would be written on the worker nodes and staged out at job termination."
                summary += "\nStageout directory for job's stdout/stderr log files:"
                summary += "\n    {}".format(stageoutLogsDir)
            else:
                summary += "\nJob's stdout/stderr log files would be written on the worker nodes, but NOT staged out at job termination."
        if logFiles:
            summary += "\nJob's stdout/stderr log filenames:"
            for f in logFiles:
                summary += "\n    {}".format(os.path.basename(f))
        elif not writeLogsOnWN:
            summary += "\nNo job's stdout/stderr log files would be created."
        summary += "\n================================================================================"
        summary += "\n===============================  end of summary  ==============================="
        summary += "\n================================================================================"

        if self.debug:
            print("Summary text created.")

        print(summary)

        #================================================================================
        # Submit the jobs
        #================================================================================

        if self.submit:
            if not self.yes:
                timeout = 5
                if self.calledFromScript:
                    print("You provided the --submit (-s) option, which instructs to submit the jobs.")
                else:
                    print("You have set the 'submit' argument to True, which instructs to submit the jobs.")
                if not self.quiet:
                    print("Please read first the summary given above, then answer whether you really want to submit the jobs.")
                    print("N.B.: You can always submit the jobs (i.e. batch scripts) yourself with the slurm 'sbatch' command.")
                question = "Are you sure you want to submit the jobs ? (you have {} minutes to answer)".format(timeout)
                signal.signal(signal.SIGALRM, questionTimeoutHandler)
                signal.alarm(timeout*60)
                try:
                    answer = interactiveYesNoQuestion(question, default='no')
                except TimeoutException as ex:
                    print(ex)
                    answer = False
                signal.alarm(0)
                if not answer:
                    print("Jobs were not submitted.")
                    return
            if useJobArray:
                print("Submitting job array consisting of {} jobs".format(numJobs))
                ret = submit(batchScripts[0], partition=sbatch_partition, qos=sbatch_qos, jobArray=True)
                if ret == 2:
                    msg = "Failed to submit jobs."
                    raise CP3SlurmUtilsException(msg)
            else:
                for j in range(numJobs):
                    print("Submitting job number {}".format(j+1))
                    ret = submit(batchScripts[j], partition=sbatch_partition, qos=sbatch_qos, jobArray=False)
                    if ret == 2:
                        msg = "Failed to submit jobs."
                        raise CP3SlurmUtilsException(msg)
        else:
            if self.calledFromScript:
                print("You have not provided the --submit (-s) option, so jobs were not submitted.")
            else:
                print("You have not set the 'submit' argument to True, so jobs were not submitted.")
            if not self.quiet:
                print("You can always submit the jobs (i.e. batch scripts) yourself with the slurm 'sbatch' command.")



from CP3SlurmUtils.Configuration import Configuration

config = Configuration()

#--------------------------------------------------------------------------------
# 1. SLURM sbatch command options
#--------------------------------------------------------------------------------

config.sbatch_partition = "cp3"
config.sbatch_qos = "cp3"
config.sbatch_chdir = "."
config.sbatch_time = "0-00:05"
config.sbatch_memPerCPU = "1024"
config.sbatch_output = "/dev/null"
config.sbatch_error = "/dev/null"
config.sbatch_additionalOptions = []

#--------------------------------------------------------------------------------
# 2. User batch script parameters that are same for all jobs
#--------------------------------------------------------------------------------

config.scratchDir = "${LOCALSCRATCH}"
config.handleScratch = False

config.environmentType = ""
config.cmsswDir = ""

config.inputSandboxContent = []
config.inputSandboxDir = ""
config.inputSandboxFilename = ""

config.batchScriptsDir = config.sbatch_chdir + "/slurm_batch_scripts"
config.batchScriptsFilename = ""

config.stageout = True
config.stageoutFiles = ["output_file_for_job_*.txt"]
# We chose the filename of the outputs to be independent of the job array id number (but dependent on the job array task id number).
# So let's put the output files in a directory whose name contains the job array id number,
# so that each job array we may submit will write in a different directory.
config.stageoutDir = config.sbatch_chdir + "/slurm_outputs/job_array_${SLURM_ARRAY_JOB_ID}"

config.writeLogsOnWN = True
config.separateStdoutStderrLogs = False
config.stdoutFilename = ""
config.stderrFilename = ""
config.stageoutLogs = True
# The default filename of the slurm logs has already a job array id number and a job array task id number in it.
# So we can put all logs together (even from different job arrays we may submit) in a unique directory; they won't overwrite each other.
config.stageoutLogsDir = config.sbatch_chdir + "/slurm_logs"

config.useJobArray = True
config.maxRunningJobs = None

# 2 jobs will be submitted, because the config parameter 'inputParams' has length 2.
config.numJobs = None

#--------------------------------------------------------------------------------
# 3 Job-specific input parameters and payload
#--------------------------------------------------------------------------------

config.inputParamsNames = ["outputFile"]

config.inputParams = [["output_file_for_job_1.txt"], ["output_file_for_job_2.txt"]]

# For job number 1, the environment variable "outputFile" will be equal to "output_file_for_job_1.txt".
# For job number 2, the environment variable "outputFile" will be equal to "output_file_for_job_2.txt".
# The payload will have access to environment variables like 'SLURM_ARRAY_JOB_ID' and 'SLURM_ARRAY_TASK_ID',
# so we can use those variables here.
config.payload = \
"""
echo "Start of 'Hello World' user payload for job ${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}"
echo "Hello World! I am a SLURM job with ID ${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}" > ${outputFile}
echo "  End of 'Hello World' user payload for job ${SLURM_ARRAY_JOB_ID}_${SLURM_ARRAY_TASK_ID}"
"""



"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import imp
import os
import pickle

from CP3SlurmUtils.Exceptions import ConfigurationException
from CP3SlurmUtils.Exceptions import CP3SlurmUtilsException


class Configuration(object):
    def __init__(self):
        object.__init__(self)
        self._internal_config_attributes = []

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if not name.startswith('_internal_'):
            if name not in self._internal_config_attributes:
                self._internal_config_attributes.append(name)
        return

    def getConfigAttrs_(self):
        return self._internal_config_attributes


def loadConfiguration(filename):
    """_loadConfiguration_
    Load a configuration file and return its Configuration attribute.
    """
    confAttr = None
    if filename.endswith(".pkl"):
        try:
            with open(filename, 'rb') as fd:
                try:
                    confAttr = pickle.load(fd)
                except Exception as ex:
                    msg = "ERROR: Failed to unpickle file {file}"
                    msg += "\nError follows:\n{error}"
                    msg = msg.format(file=filename, error=str(ex))
                    raise CP3SlurmUtilsException(msg)
        except IOError as ex:
            msg = "ERROR: Failed to open file {file}"
            msg += "\nError follows:\n{error}"
            msg = msg.format(file=filename, error=str(ex))
            raise CP3SlurmUtilsException(msg)
    else:         
        cfgBaseName = os.path.basename(filename).replace(".py", "")
        cfgDirName = os.path.dirname(filename)
        if not cfgDirName:
            modPath = imp.find_module(cfgBaseName)
        else:
            modPath = imp.find_module(cfgBaseName, [cfgDirName])
        try:
            modRef = imp.load_module(cfgBaseName, modPath[0], modPath[1], modPath[2])
        except Exception as ex:
            msg = "ERROR: Unable to load configuration file {file}"
            msg += "\nError follows:\n{error}"
            msg = msg.format(file=filename, error=str(ex))
            raise CP3SlurmUtilsException(msg)
        for attr in modRef.__dict__.values():
            if isinstance(attr, Configuration):
                if confAttr is None:
                    confAttr = attr
                else:
                    msg = "ERROR: Configuration file contains more than one Configuration object."
                    raise ConfigurationException(msg)
    return confAttr



"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import string
import collections


def words(text):
    return re.findall('[a-z]+', text.lower())


def train(features):
    model = collections.defaultdict(lambda: 1)
    for f in features:
        model[f] += 1
    return model


DICTIONARY = []


def edits1(word):
    splits     = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes    = [a + b[1:] for a, b in splits if b]
    transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b)>1]
    replaces   = [a + c + b[1:] for a, b in splits for c in string.ascii_lowercase if b]
    inserts    = [a + c + b     for a, b in splits for c in string.ascii_lowercase]
    return set(deletes + transposes + replaces + inserts)


def known_edits2(word):
    return set(e2 for e1 in edits1(word) for e2 in edits1(e1) if e2 in DICTIONARY)


def known(words):
    return set(w for w in words if w in DICTIONARY)


def correct(word):
    candidates = known([word]) or known(edits1(word)) or known_edits2(word) or [word]
    return max(candidates, key=DICTIONARY.get)


def is_correct(word):
    return word in DICTIONARY



"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

class CP3SlurmUtilsException(Exception):
    """ General exception """
    pass


class ConfigurationException(CP3SlurmUtilsException):
    pass



"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import datetime
import os
import subprocess


def submit(batchScript, partition=None, qos=None, wckey=None, jobArray=False, bookkeep=True):
    timestamp = str(datetime.datetime.now().strftime("%Y%m%d-%H%M%S%f"))
    submitOpts = ""
    if partition:
        submitOpts += " --partition={}".format(partition)
    if qos:
        submitOpts += " --qos={}".format(qos)
    if wckey:
        submitOpts += " --wckey={}".format(wckey)
    submitCmd = "sbatch{} {}".format(submitOpts, batchScript)
    print("Submission command:\n    {}".format(submitCmd))
    process = subprocess.Popen(submitCmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    stdout = stdout.decode('utf-8')
    stderr = stderr.decode('utf-8')
    returncode = process.returncode
    if returncode != 0:
        msg = "ERROR: Failed to submit."
        if stdout:
            msg += "\nStdout:\n\t{}".format(str(stdout).replace('\n', '\n\t'))
        if stderr:
            msg += "\nStderr:\n\t{}".format(str(stderr).replace('\n', '\n\t'))
        print(msg)
        return 2
    stdout = stdout.replace('\n', '')
    print(stdout)
    if not bookkeep:
        return 0
    jobId = ''
    if stdout.startswith('Submitted batch job '):
        jobId = stdout.replace('Submitted batch job ', '')
    if not jobId:
        msg = "WARNING: Failed to parse the sbatch stdout to retrieve the job id."
        msg += " Will not write the job to the user's slurm_jobs file."
        print(msg)
        return 1
    filename = os.path.expanduser('~/.slurm_jobs')
    try:
        with open(filename, 'a+') as fd:
            fd.write('{} {} {} jobarray={}\n'.format(timestamp, jobId, batchScript, 'T' if jobArray else 'F'))
    except IOError as ex:
        msg = "WARNING: Failed to write job to user's slurm_jobs file {file}"
        msg += "\nError follows:\n{error}"
        msg = msg.format(file=filename, error=str(ex))
        print(msg)
        return 1
    try:
        newFileContent = ""
        with open(filename, 'r') as fd:
            for line in fd:
                date, time = line.split()[0].split('-')
                year, month, day = date[:4], date[4:6], date[6:8]
                hour, minute, second, microsecond = time[:2], time[2:4], time[4:6], time[6:12]
                lineDateTime = datetime.datetime(int(year), int(month), int(day), int(hour), int(minute), int(second), int(microsecond))
                now = datetime.datetime.now()
                if (now - lineDateTime).days < 30:
                    newFileContent += line
        with open(filename, 'w') as fd:
            fd.write(newFileContent)
    except Exception as ex:
        msg = "WARNING: Failed to clean user's slurm_jobs file {file}"
        msg += "\nError follows:\n{error}"
        msg = msg.format(file=filename, error=str(ex))
        print(msg)
        return 1
    return 0



""" Setuptools-based setup module for CP3SlurmUtils """

from setuptools import setup, find_packages
import os, os.path

from io import open
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

pkgName = "CP3SlurmUtils"

setup(
    name=pkgName,
    version="0.2.0",

    description="Utilities package to submit jobs to SLURM using as input a python sbatch configuration file.",
    long_description=long_description,
    long_description_content_type='text/markdown',

    url="https://gitlab.cern.ch/cp3-cms/CP3SlurmUtils",

    author="Andres Tanasijczuk (Université catholique de Louvain)",
    author_email="cp3-support@listes.uclouvain.be",
    license="GPL-3.0-or-later",

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Distributed Computing',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='slurm',

    package_dir={'' : os.path.join('src', 'python')},
    packages=find_packages(os.path.join('src', 'python')),
    scripts=[os.path.join(root, item) for root, subFolder, files in os.walk('bin') for item in files],
    package_data={pkgName: ['examples/*']},
    data_files=[(os.path.join('etc', pkgName), [os.path.join('etc', 'defaults.cfg.example')])],
)



"""
Copyright (C) 2019  Universite catholique de Louvain, Belgium.

This file is part of CP3SlurmUtils.

CP3SlurmUtils is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CP3SlurmUtils is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CP3SlurmUtils.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys


def interactiveYesNoQuestion(question, default='no', answerYes=False):
    """Ask a yes/no question and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).
    The "answer" return value is True for "yes" or False for "no".
    """
    validAnswers = {'yes': True, 'y': True, 'no': False, 'n': False}
    if answerYes:
        return validAnswers['yes']
    prompt = " [y/n] "
    if default in ['yes', 'y']:
        prompt = " [Y/n] "
    if default in ['no', 'n']:
        prompt = " [y/N] "
    while True:
        sys.stdout.write(question + prompt)
        try:
            # Python 2
            _input = raw_input
        except NameError:
            # Python 3
            _input = input
        choice = _input().lower()
        if choice in validAnswers:
            return validAnswers[choice]
        elif choice == '' and default is not None:
            return validAnswers[default]
        else:
            sys.stdout.write("Please respond with 'yes' ('y') or 'no' ('n').\n")


class TimeoutException(Exception):
    pass


def questionTimeoutHandler(signum, frame):
    msg = "\nSorry, the timeout limit to answer the question has been reached."
    raise TimeoutException(msg)
